//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/corr/corr.c,v $
// $Id: corr.c,v 1.2 2003/03/11 22:51:33 ah Exp $ 
//
// Description: print information about specified wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: corr.c,v 1.2 2003/03/11 22:51:33 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <stdlib.h>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
USING_NAMESPACE(std)

#include <AH/Config/Types.h>
#include <AH/Etech/Types.h>
#include <AH/Etech/Riaa.h>
#include <AH/Etech/Algo.h>
#include <AH/IO/Riff/WaveFile.h>
#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/Error/Sys.h>
USING_NAMESPACE(AH)

// --- locaL definitions ---

// --- local constants ---

static const char* cmdLineOptions = "Cc:dhnr:v";
static const char* version = "0.5.1";

// --- local data ---

static unsigned int debon = 0;
static bool normalize = false;
static bool cyclic = false;
static unsigned int channel = 1;

//static char mode = '?';

// --- local classes ---

// --- local functions ---


static void usage()
{
    cerr << "Usage: corr [-" << cmdLineOptions << 
	"] inputfile [inputfile] outputfile\n"
	 << "       make an auto or cross correlation.\n"
	 << "  options:\n"
	 << "    h         - print this help message\n"
	 << "    C         - cyclic correlation\n"
	 << "    c <num>   - channel number (for multi channel WAV files)\n"
	 << "                stereo: 1=left 2=right (default=1)\n"
	 << "    d         - increase debug output (may be given more than once)\n"
	 << "    n         - normalize to highest point = 1 (not for WAV)\n"
	 << "    v         - print version\n"
	 << "  input file  - WAV file or 16 bit mono PCM\n"
	 << "  output file - text file for gnuplot or mono WAV file" << endl;
} // end of usage()



static void print_version()
{
    cerr << "corr Vs " << version << " " 
	 << "(c) 2002 Andreas Hnnebeck\n"
	 << "Credits to Patrick Piecha and Thomas Ludwig.\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()



static unsigned int read_pcm_file(double* data, unsigned int num, FILE* fpi)
{
    if (debon > 1) cerr << "read_file: read " << num << " points" << endl;
    unsigned int i = 0;

    int ret = 0;
    do
    {
	const unsigned int snum = 65536;
	int16 sbuf[snum];

	// set size of data to read on this turn:
	unsigned int rnum = snum;
	if (rnum > num) rnum = num;

	// read input buffer:
	if (debon > 2) cerr << "read_file: read " << rnum << " points" << endl;
	ret = fread(sbuf, sizeof(int16), rnum, fpi);
	if (debon > 2) cerr << "read_file: read returns " << ret << endl;
	if (-1 == ret) throw ErrorSys();
	for (int k = 0; k < ret; ++k)
	    data[i++] = sbuf[k];
	num -= ret;
	if ((unsigned int)ret < rnum) break;
    } while (num > 0);

    if (debon > 1) cerr << "read_file: returns " << i << endl;
    return i; // number of read entries
}



static uint32 read_wav_file(double* data, uint32 num, IORiffWaveFile& wave)
{
    if (debon > 1) cerr << "read_file: read " << num << " points" << endl;
    uint32 i = 0;


    // get size of data for desired samples
    const uint32 snum = 16384; // number of samples to read
    const uint32 bufsize = wave.getDataBufSize(snum);

    int ret = 0;
    do
    {
	unsigned char sbuf[bufsize];
	memset(sbuf, 0, bufsize);

	// set size of data to read on this turn:
	uint32 rnum = snum;
	uint32 rbufsize = bufsize;
	if (rnum > num)
	{
	    rnum = num;
	    rbufsize = wave.getDataBufSize(num);
	}

	// read input buffer:
	if (debon > 2) cerr << "read_file: read " << rnum << " points" << endl;
	ret = wave.readRaw(sbuf, rbufsize);
	if (ret <= 0) break;
	if ((uint32)ret < rbufsize)
	    cerr << "Warning: could not read " << (rbufsize - ret) 
		 << " bytes from WAV file" << endl;  
	for (uint32 k = 0; k < rnum; ++k)
	{
	    data[i++] = wave.getBufSample(sbuf, k, channel);
//	    data[i++] = (wave.getBufSample(sbuf, k, channel) > 0 ? 1 : -1);
	}
	num -= rnum;
    } while (num > 0);

    if (debon > 1) cerr << "read_file: returns " << i << endl;
    return i; // number of read entries
}



bool isWaveFile(const std::string& filename)
{
    const char* fn = filename.c_str();
    int len = strlen(fn);
    if (0 == strcasecmp(".wav", fn + len - 4))
	return true;
    return false;
}



// get normalization factor 
static double getNormFact(uint32 num, const double* data)
{
    if (!normalize) return 1.0;

    // normalize at highest point
    if (debon) cerr << "normalize data at highest point" << endl;
    double max = 0;
    double min = 0;
    for (unsigned int i = 0; i < num; ++i)
    {
	double val = data[i];
	if (val > max) max = val;
	if (val < min) min = val;
    }
    min = -min;
    if (min > max) max = min;
    if (max == 0) return 1.0;

    double norm_fact = 1.0 / max;
    if (debon > 1) 
	cerr << "normalize data : peak = " << max << " norm_fact = "
	     << norm_fact << endl;
    return norm_fact;
}


static void correlate(const double* inbuf, const size_t inbuflen, 
		      const double* coeffs, const size_t coeffslen,
		      double* outbuf, const size_t outbuflen) 
{
    /* clear output-buffer before processing */
    memset (outbuf, 0, outbuflen * sizeof (*outbuf));

#if 1
    size_t kmax = inbuflen + coeffslen;
    for (size_t k = 0; k < kmax; ++k ) 
    {
	if (debon && (0 == k % 50))
	{
	    cerr << k << "/ "
		 << (int)((double)k / (double)kmax * 100 + 0.5) 
		 << "%\r";
	}
	for (size_t n = 0; n < inbuflen; ++n )
	    outbuf[k % outbuflen] += inbuf[n] * coeffs[(k+n) % coeffslen];
    }
    if (debon) cerr << outbuflen << "\n";
#else
    /* convolve buffers (fast version) */
    size_t   n;
    size_t   k;

    for ( n = 0; n < inbuflen; n++ ) {
	for ( k = 0; k < coeffslen; k++ ) {
	    // cyclic!
	    outbuf[(n + k) % inbuflen] += inbuf[n] * coeffs[coeffslen - k - 1];
	}
    }
#endif
}

static void calc_stat(const std::string&  head, double* data, unsigned int num)
{
    // determine min, mean and max value:
    double mean = 0;
    double max = 0;
    double min = 0;
    for (unsigned int i = 0; i < num; ++i)
    {
	mean += data[i];
	if (data[i] > max) max = data[i];
	if (data[i] < min) min = data[i];
    }
    mean /= num;
    cerr << head << ": min = " << min << ", mean = " << mean 
	 << ", max = " << max << "; num=" << num << endl;
}



static double* read_file
(
    const string& in,
    uint32& num,      // number of samples
    uint32& sr,       // sample rate in Hz
    uint32& sb        // number of sample bits
)
{
    //------------------------------------
    // read the data from the input file:
    //------------------------------------

    double* din = 0;      // input samples (double)
    if (isWaveFile(in))
    {
	IORiffWaveFile wave(in.c_str());
	if (debon > 1) wave.printInfo(cerr);
	uint32 chnum = wave.getChannelNum();
	if (channel > chnum) 
	    throw ErrorBasic("illegal argument following option -c");

	num = wave.getSampleNum();
	sb = wave.getSampleSize();
	uint32 sr_in = wave.getSampleRate();

	if (0 == sr) sr = sr_in;
	else if (sr != sr_in) throw ErrorBasic("non matching samplerate");

	// create input buffer
	din = new double[num];
	if (!din) throw ErrorSysOutOfMem();

	// read input data into input buffer:
	if (debon) cerr << "Input: reading wav file " << in << endl;
	uint32 gotten = read_wav_file(din, num, wave);
	if (((int)num - (int)gotten) > 0)
	{
	    cerr << "WARNING: " << (int)num - (int)gotten 
		 << " missing data points in " << in << endl;
	    num = gotten;
	}
    }
    else // binary PCM file
    {
	// get size of file:
	struct stat st;
	int ret = stat(in.c_str(), &st);
	if (0 != ret) throw ErrorSys("Cannot stat " + in);

	// determine number of data points:
	num = st.st_size / sizeof(short);
	if (debon > 2) cerr << in << ": size=" << st.st_size << " = " << num
			    << " points" << endl;
	if (2 > num) throw ErrorBasic("Size of " + in + " is too small");

	// open data file for read:
	FILE* fpi = fopen(in.c_str(), "rb");
	if (!fpi) throw ErrorSys("Cannot open " + in);

	// create input buffer
	din = new double[num];
	if (!din)
	{
	    fclose(fpi);
	    throw ErrorSysOutOfMem();
	}

	// read input data into input buffer:
	if (debon) cerr << "Input: reading PCM file " << in << endl;
	uint32 gotten = read_pcm_file(din, num, fpi);

	// close file, done:
	fclose(fpi);
	if (((int)num - (int)gotten) > 0)
	{
	    cerr << "WARNING: " << (int)num - (int)gotten 
		 << " missing data points in " << in << endl;
	    num = gotten;
	}

    }
    if (debon) calc_stat (in, din, num);
    return din;
}


static void write_file(const string& out, uint32 sr, uint32 sb, 
		       double* data, uint32 num)
{
    if (isWaveFile(out))
    {
	// create WAV file:
	int channels = 1;
	IORiffWaveFile wave(out.c_str(), channels, sb, sr, num);
    
	// get size of data for desired samples
	const uint32 snum = 65536; // number of samples to write
	const uint32 bufsize = wave.getDataBufSize(snum);

	double min = 0;
	double max = 0;
	for (unsigned int k = 0; k < num; ++k)
	{
	    double cur = data[k];
	    if (cur > max) max = cur;
	    if (cur < min) min = cur;
	}
	min = -min;
	if (max < min) max = min;
	double norm = 32000/max;

	uint32 osize = num;
	uint32 i = 0;
	int ret = 0;
	do
	{
	    unsigned char sbuf[bufsize];
	    memset(sbuf, 0, bufsize);

	    // set size of data to read on this turn:
	    uint32 rnum = snum;
	    uint32 rbufsize = bufsize;
	    if (rnum > osize)
	    {
		rnum = osize;
		rbufsize = wave.getDataBufSize(osize);
	    }

	    // write double data (dithered) into output buffer
	    for (uint32 k = 0; k < rnum; ++k)
	    {
		double db_mean = norm * data[i];
		int32 val = (int32)(db_mean + 0.5);
		if (debon > 3)
		    cerr << "write_file: val = " << val << endl;
		for (int j = 0; j < channels; ++j)
		    wave.setBufSample(sbuf, k, j+1, val);
		++i;
	    }

	    // write output buffer:
	    if (debon > 2) 
		cerr << "write_file: write " << rnum << " points" << endl;
	    ret = wave.writeRaw(sbuf, rbufsize);
	    if (ret <= 0) break;
	    if ((uint32)ret < rbufsize)
		cerr << "Warning: could not write " << (rbufsize - ret) 
		     << " bytes to WAV file" << endl;  
	    osize -= rnum;
	} while (osize > 0);

	wave.writeHeader();
    }
    else
    {
	// get normalisation factor if desired:
	double nf = getNormFact(num, data);

	// open output file
	FILE* fpo = fopen(out.c_str(), "w");
	if (!fpo) throw ErrorSys("Cannot open " + out);

	// print comment line:
	fprintf(fpo, "# num | xcorr\n");
	// print entries:
	for (unsigned int i = 0; i < num; ++i)
	    fprintf(fpo, "%d\t%f\n", i, data[i] * nf);
	fclose(fpo);
    }
}


static void process(const string& in1, const string& in2, const string& out)
{
    bool autocorr = (in1 == in2);

    //---------------------------------------
    // read the data from the 1. input file:
    //---------------------------------------
    uint32 num1 = 0;  // number of samples in 1. file
    uint32 sb1 = 0;   // number of bits in sample
    uint32 sr = 0;       // sample rate in Hz
    double* data1 = read_file(in1, num1, sr, sb1);

    //---------------------------------------
    // read the data from the 2. input file:
    //---------------------------------------
    uint32 num2 = 0;  // number of samples in 2. file
    uint32 sb2 = 0;   // number of bits in sample
    double* data2;
    if (!autocorr)
    {
	data2 = read_file(in2, num2, sr, sb2);
    }
    else
    {
	data2 = data1;
	num2 = num1;
	sb2 = sb1;
    }

    // use higher sample bit number for output file:
    uint32 sb = (sb1 > sb2 ? sb1 : sb2);

    //-------------------------------------------------------
    // calculate the cross correlation from the input data:
    //-------------------------------------------------------
    if (debon) cerr << "executing " << (autocorr ? "auto" : "cross")
		    << " correlation" << endl;
    uint32 num = num1 + num2;        // number of correlated data
    if (cyclic) num = num1;

    double* dout = new double[num];
    if (!dout)
    {
	delete[] data1;
	if (!autocorr) delete[] data2;
	throw ErrorSysOutOfMem();
    }
    correlate(data1, num1, data2, num2, dout, num);
    delete[] data1;
    if (!autocorr) delete[] data2;
    if (debon) calc_stat ("xcorr", dout, num);

    //-----------------------------------------
    // write the results into the output file:
    //-----------------------------------------
    if (debon) cerr << "writing file " << out << endl;

    write_file(out, sr, sb, dout, num);
    delete dout;
}


// --- global functions ---


int main(int argc, char * const argv[])
{
    // loop through command line options:
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'd': ++debon; break;
	case 'C': cyclic = true; break;
	case 'c': channel = atoi(optarg); break;
	case 'h': usage(); exit(0);
	case 'n': normalize = true; break;
	case 'v': print_version(); exit(0);
	default : usage(); exit(1);
	}
    }
    if (optind >= argc) 
    {
	cerr << "Error: missing file" << endl;
	usage();
	exit(1);
    }

    // loop through the remaining arguments: these are files
    vector<string> args;
    for (int i=optind; i < argc; i++)
    {
	args.push_back(argv[i]);
    }

    string in_file_1;
    string in_file_2;
    string out_file;
    switch (args.size())
    {
    case 3:  in_file_1 = args[0];
	     in_file_2 = args[1];
	     out_file = args[2];
	     break;
    case 2:  in_file_1 = in_file_2 = args[0];
	     out_file = args[1];
	     break;
    case 1:  cerr << "Error: missing input file(s)" << endl;
	     usage();
	     exit(1);
	     break;
    case 0:  cerr << "Error: missing input and output files" << endl;
	     usage();
	     exit(1);
    default: cerr << "Error: too many arguments" << endl;
	     usage();
	     exit(1);
    }

    // process the files
    static unsigned int errnum = 0;
    try
    {
	process(in_file_1, in_file_2, out_file); 
    }
    catch (ErrorClass& err)
    {
	++errnum;
	cerr << "Error: " << err.getMsg() << endl;
    }
    catch (exception& err)
    {
	++errnum;
	cerr << "Error: " << err.what() << endl;
    }
    
    exit (errnum ? 1 : 0);
    return  (errnum ? 1 : 0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
