//---------------------------------------------------------------------------
// Copyright 2005 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/libAHio/CddaSector.c,v $
// $Id: CddaSector.c,v 1.1 2005/01/11 00:28:46 ah Exp $ 
//
// Description: calculate TOC for cdrdao
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: CddaSector.c,v 1.1 2005/01/11 00:28:46 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
#include <AH/Config/Types.h>
#include <AH/IO/Cdda/Sector.h>
#include <AH/STL/String.h>
#include <AH/Error/Basic.h>
USING_NAMESPACE(AH)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


// --- local constants ---

// --- local variables ---

// --- local classes ---

// --- local functions ---

// --- global functions ---

BEGIN_NAMESPACE(AH)

CddaSector::CddaSector(double time)
{
    this->minutes = (uint32)(time / 60.0);
    double seco = time - (this->minutes * 60);
    this->seconds = (uint32)seco;
    double sect = (seco - this->seconds) * SECTORS_PER_SECOND;
    this->sectors = (uint32)sect;
    // ignore remaining stuff

    this->sanityCheck("CddaSector time constructor");
}


CddaSector::CddaSector(uint32 bytes)
{
    this->minutes = bytes / BYTES_PER_MINUTE;
    bytes %= BYTES_PER_MINUTE;
    this->seconds = bytes / BYTES_PER_SECOND;
    bytes %= BYTES_PER_SECOND;
    this->sectors = bytes / BYTES_PER_SECTOR;
    // ignore remaining bytes
//    uint32 rem = bytes % BYTES_PER_SECTOR;

    this->sanityCheck("CddaSector byte constructor");
}

// convert string into CddaSector
CddaSector::CddaSector(const std::string& s)
{
    std::vector<std::string> token;
    STLStringToken(s, token, ":");
    if (token.size() != 3) 
	throw ErrorBasic(s + ": illegal CDDA sector");

    int minu = atoi(token[0].c_str());
    int seco = atoi(token[1].c_str());
    int sect = atoi(token[2].c_str());
    if ((minu < 0) || (minu > 99) ||
	(seco < 0) || (seco > 59) ||
	(sect < 0) || (sect > ((int)SECTORS_PER_SECOND - 1)))
	throw ErrorBasic(s + ": illegal CDDA sector");

    this->minutes = (uint32)minu;
    this->seconds = (uint32)seco;
    this->sectors = (uint32)sect;
}


void CddaSector::sanityCheck(const std::string& msg) const
{
    bool sanity = true;
    if (this->minutes > 99) sanity = false;
    if (this->seconds > 59) sanity = false;
    if (this->sectors > (SECTORS_PER_SECOND - 1)) sanity = false;
    if (!sanity) throw ErrorBasic(msg + ": bad sector value");
}

// return value as string
std::string CddaSector::asString() const
{
    char buf[64];
    (void)sprintf(buf, "%02d:%02d:%02d", minutes, seconds, sectors);
    return buf;
}

// return value as bytes
uint32 CddaSector::asBytes() const
{
    uint32 sect = this->sectors;
    sect += this->seconds * SECTORS_PER_SECOND;
    sect += this->minutes * 60 * SECTORS_PER_SECOND;
    return sect * BYTES_PER_SECTOR;
}

// return value as time (seconds)
double CddaSector::asTime() const
{
    double seco = this->minutes * 60;
    seco += this->seconds;
    seco += this->sectors / SECTORS_PER_SECOND;
    return seco;
}

// overloaded operator =
CddaSector& CddaSector::operator = (const CddaSector &cs)
{
    if (this != &cs)
    {
	this->minutes = cs.minutes;
	this->seconds = cs.seconds;
	this->sectors = cs.sectors;
    }
    return *this;
}

// overloaded operator += for addition of two CDDA sector values
CddaSector& CddaSector::operator += (const CddaSector &cs2)
{
    uint32 bytes = cs2.asBytes();
    bytes += this->asBytes();
    CddaSector cs(bytes);
    *this = cs;
    return *this;
}

// overloaded operator -= for addition of two CDDA sector values
CddaSector& CddaSector::operator -= (const CddaSector &cs2)
{
    uint32 b1 = this->asBytes();
    uint32 b2 = cs2.asBytes();
    if (b1 > b2)
    {
	CddaSector cs(b1 - b2);
	*this = cs;
	return *this;
    }
    throw ErrorBasic("CddaSector subtraction failed");
}

// overloaded operator * for multiplication of CddaSector with double
CddaSector& CddaSector::operator *= (double k)
{
    uint32 bytes = (uint32)(double(this->asBytes()) * k);
    CddaSector csk(bytes);
    *this = csk;
    return *this;
}

// overloaded operator / for division of CddaSector by double
CddaSector& CddaSector::operator /= (double k)
{
    uint32 bytes = (uint32)(double(this->asBytes()) / k);
    CddaSector csk(bytes);
    *this = csk;
    return *this;
}

// overloaded operator + for addition of two CDDA sector values.
CddaSector operator + (const CddaSector &cs1, const CddaSector &cs2)
{
    uint32 bytes = cs1.asBytes();
    bytes += cs2.asBytes();
    CddaSector cs(bytes);
    return cs;
}

// overloaded operator - for subtraction of two CDDA sector values.
CddaSector operator - (const CddaSector &cs1, const CddaSector &cs2)
{
    uint32 b1 = cs1.asBytes();
    uint32 b2 = cs2.asBytes();
    if (b1 > b2)
    {
	CddaSector cs(b1 - b2);
	return cs;
    }
    throw ErrorBasic("CddaSector subtraction failed");
}

// overloaded operator * for multiplication of CddaSector with double.
CddaSector operator * (const CddaSector &cs, double k)
{
    uint32 bytes = (uint32)(k * double(cs.asBytes()));
    CddaSector csk(bytes);
    return csk;
}

// overloaded operator * for multiplication of double with CddaSector.
CddaSector operator * (double k, const CddaSector &cs)
{
    uint32 bytes = (uint32)(k * double(cs.asBytes()));
    CddaSector csk(bytes);
    return csk;
}

// overloaded operator / for division of CddaSector by double.
CddaSector operator / (const CddaSector &cs, double k)
{
    uint32 bytes = (uint32)(double(cs.asBytes()) / k);
    CddaSector csk(bytes);
    return csk;
}

// overloaded operator == for comparison of two CddaSectors.
bool operator == (const CddaSector &cs1, const CddaSector& cs2)
{
    return (cs1.minutes == cs2.minutes) &&
	(cs1.seconds == cs2.seconds) &&
	(cs1.sectors == cs2.sectors);
}

// overloaded operator != for comparison of two CddaSectors.
bool operator != (const CddaSector &cs1, const CddaSector& cs2)
{
    return (cs1.minutes != cs2.minutes) ||
	(cs1.seconds != cs2.seconds) ||
	(cs1.sectors != cs2.sectors);
}

// overloaded operator < for comparison of two CddaSectors.
bool operator < (const CddaSector &cs1, const CddaSector& cs2)
{
    return cs1.asBytes() < cs2.asBytes();
}

// overloaded operator > for comparison of two CddaSectors.
bool operator > (const CddaSector &cs1, const CddaSector& cs2)
{
    return cs1.asBytes() > cs2.asBytes();
}

END_NAMESPACE

// --- end of file ---
